#include "BME280.h"
#include <wiringPi.h>
#include <wiringPiI2C.h>


//----------------------------------------------
// function
//    コンストラクタ
// parameter
//    なし
// return
//    なし
//----------------------------------------------
CBME280::CBME280()
{

}

//----------------------------------------------
//	function
//		デストラクタ
//	parameter
//		なし
//	return
//		なし
//----------------------------------------------
CBME280::~CBME280()
{

}

//----------------------------------------------
//	function
//		I2Cの初期化
//		補正値の取得
//	parameter
//		int     iDevId	    [in]デバイスID(i2cdetectで見つかる)
//      LPSTR   szBusName   [in]Bus名(i2cdetect -lで表示される名前)/NULL
//	return
//		TRUE/FALSE
//----------------------------------------------
BOOL CBME280::Setup(int iDevId, LPSTR pszBusName)
{
    BYTE    osrs_t = 1;             // Temperature oversampling x 1
    BYTE    osrs_p = 1;             // Pressure oversampling x 1
    BYTE    osrs_h = 1;             // Humidity oversampling x 1
    BYTE    mode = 3;               // Normal mode
    BYTE    t_sb = 5;               // Tstandby 1000ms
    BYTE    filter = 0;             // Filter off 
    BYTE    spi3w_en = 0;           // 3-wire SPI Disable
    BYTE    ctrl_meas_reg = (osrs_t << 5) | (osrs_p << 2) | mode;
    BYTE    config_reg = (t_sb << 5) | (filter << 2) | spi3w_en;
    BYTE    ctrl_hum_reg = osrs_h;
    BOOL    fRet = FALSE;

    m_iTfine = 0;
// http://www.ne.jp/asahi/o-family/extdisk/BME280/BME280_DJP.pdf
//  バスを指定できる関数を使う方が良いでしょう(i2cdetect -lで表示される名前)
    if (pszBusName == NULL)
    {
        if ((m_fdI2C = wiringPiI2CSetup(iDevId)) == -1)
            goto L_END;
    }
    else
    {
        if ((m_fdI2C = wiringPiI2CSetupInterface(pszBusName, iDevId)) == -1)
            goto L_END;
    }
    if (writeReg(0xF2, ctrl_hum_reg) == FALSE)      // 7.4.3
        goto L_END;
    if(writeReg(0xF4, ctrl_meas_reg) == FALSE)      // 7.4.5
        goto L_END;
    if(writeReg(0xF5, config_reg) == FALSE)         // 7.4.6
        goto L_END;
    readTrim();                                     // 調整パラメーターの読み出し(6.2.2)

    fRet = TRUE;
L_END:
    return(fRet);

}

//----------------------------------------------
//	function
//		温度の取得(C)
//		レジスタのアドレス　0xFA 0xFB 0xFC
//	parameter
//		なし
//	return
//		温度(C)
//----------------------------------------------
double CBME280::readTemp()
{
    BYTE    data[3];
    int     temp_raw;

    readRegs(0xFA, 3, data);
    temp_raw = (data[0] << 12) | (data[1] << 4) | ((data[2] >> 4) & 0x0f);
    return((double)calibration_T(temp_raw)/100.0);
}

//----------------------------------------------
//	function
//		気圧の取得(hPa)
//		レジスタのアドレス　0xF7 0xF8 0xF9
//	parameter
//		なし
//	return
//		気圧(hPa)
//----------------------------------------------
double CBME280::readPres()
{
    BYTE    data[3];
    int     pres_raw;

    readRegs(0xF7, 3, data);

    pres_raw = (data[0] << 12) | (data[1] << 4) | ((data[2] >> 4) & 0x0f);
    return((double)calibration_P(pres_raw) / (256.0 * 100.0));
}

//----------------------------------------------
//	function
//		湿度の取得(%)
//		レジスタのアドレス　0xFD 0xFE
//	parameter
//		なし
//	return
//		湿度(%)
//----------------------------------------------
double CBME280::readHum()
{
    BYTE    data[2];
    int     pres_raw;

    readRegs(0xFD, 2, data);
    pres_raw = (data[0] << 8) | data[1];

    return((double)calibration_H(pres_raw) / 1024.0);
}

//----------------------------------------------
//	function
//		温度の値を補正(6.2.3)
//	parameter
//		int adc_T       [in]レジスタから読み込んだ値
//	return
//		温度(C)*100
//----------------------------------------------
int CBME280::calibration_T(int adc_T)
{
    int     var1, var2, T;

    var1 = ((((adc_T >> 3) - ((int)m_uDig_T1 << 1))) * m_iDig_T2) >> 11;
    var2 = (((((adc_T >> 4) - ((int)m_uDig_T1)) * ((adc_T >> 4) - ((int)m_uDig_T1))) >> 12) * m_iDig_T3) >> 14;

    m_iTfine = var1 + var2;
    T = (m_iTfine * 5 + 128) >> 8;
    return(T);
}

//----------------------------------------------
//	function
//		気圧の値を補正
//	parameter
//		int adc_P       [in]レジスタから読み込んだ値
//	return
//		気圧(Pa)*256
//----------------------------------------------
UINT CBME280::calibration_P(int adc_P)
{
    SInt64  var1, var2, p;

    var1 = ((SInt64)m_iTfine) - 128000;
    var2 = var1 * var1 * (SInt64)m_iDig_P6;
    var2 = var2 + ((var1 * (SInt64)m_iDig_P5) << 17);
    var2 = var2 + (((SInt64)m_iDig_P4) << 35);
    var1 = ((var1 * var1 * (SInt64)m_iDig_P3) >> 8) + ((var1 * (SInt64)m_iDig_P2) << 12);
    var1 = (((((SInt64)1) << 47) + var1)) * ((SInt64)m_uDig_P1) >> 33;
    if (var1 == 0)
    {
        return 0; // ゼロ除算による例外を避ける。
    }
    p = 1048576 - adc_P;
    p = (((p << 31) - var2) * 3125) / var1;
    var1 = (((SInt64)m_iDig_P9) * (p >> 13) * (p >> 13)) >> 25;
    var2 = (((SInt64)m_iDig_P8) * p) >> 19;
    p = ((p + var1 + var2) >> 8) + (((SInt64)m_iDig_P7) << 4);
    return (UINT)p;
}

//----------------------------------------------
//	function
//		湿度の値を補正
//	parameter
//		int adc_H       [in]レジスタから読み込んだ値
//	return
//		湿度(%)*1024
//----------------------------------------------
UINT CBME280::calibration_H(int adc_H)
{
    int     v_x1_u32r;
    v_x1_u32r = (m_iTfine - ((int)76800));
    v_x1_u32r = (((((adc_H << 14) - (m_iDig_H4 << 20) - (m_iDig_H5 * v_x1_u32r)) + ((int)16384)) >> 15) * (((((((v_x1_u32r * m_iDig_H6) >> 10) * 
                (((v_x1_u32r * ((int)m_bDig_H3)) >> 11) + ((int)32768))) >> 10) + ((int)2097152)) *
                m_iDig_H2 + 8192) >> 14));
    v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((int)m_bDig_H1)) >> 4));
    v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
    v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);
    return (UINT)(v_x1_u32r >> 12);
}

//----------------------------------------------
//	function
//		8bitデータをレジスタに書き込む
//	parameter
//      BYTE reg_address    [in]レジスタ
//      BYTE data           [in]データ
//	return
//		TRUE/FALSE
//----------------------------------------------
BOOL CBME280::writeReg(BYTE reg_address, BYTE data)
{
    BOOL    fRet = TRUE;

    if (wiringPiI2CWriteReg8(m_fdI2C, reg_address, data) < 0)   // 8 bit data value into the device register indicated.
    {
        DispNetErrorMsg("Err:wiringPiI2CWriteReg8");
        fRet = FALSE;
    }
    return(fRet);
}

//----------------------------------------------
//	function
//		8bitデータをレジスタから指定した数読み込む
//	parameter
//      BYTE    reg_address [in]レジスタ
//      int     iNum        [in]個数
//      BYTE    *data       [in/out]データ
//	return
//		なし
//----------------------------------------------
void CBME280::readRegs(BYTE reg_address, int iNum, BYTE *data)
{
    int     i;

    for (i = 0; i < iNum; i++) 
    {
        *data = wiringPiI2CReadReg8(m_fdI2C, reg_address);      // 8 value from the device register indicated
        reg_address++;
        data++;
    }
}


//----------------------------------------------
//	function
//		調整パラメーターの読み出し(6.2.2)
//	parameter
//      なし
//	return
//		TRUE/FALSE
//----------------------------------------------
void CBME280::readTrim()
{
    BYTE data[32], i = 0;

    readRegs(0x88, 24, data);           // dig_T1-dig_P9
    i += 24;
    readRegs(0xA1, 1, data + i);        // dig_H1
    i += 1;
    readRegs(0xE1, 7, data + i);        // dig_H2-dig_H6
    i += 7;

    m_uDig_T1 = (data[1] << 8) | data[0];				// 符号なし
    m_iDig_T2 = (short)((data[3] << 8) | data[2]);
    m_iDig_T3 = (short)((data[5] << 8) | data[4]);

    m_uDig_P1 = (data[7] << 8) | data[6];				// 符号なし
    m_iDig_P2 = (short)((data[9] << 8) | data[8]);
    m_iDig_P3 = (short)((data[11] << 8) | data[10]);
    m_iDig_P4 = (short)((data[13] << 8) | data[12]);
    m_iDig_P5 = (short)((data[15] << 8) | data[14]);
    m_iDig_P6 = (short)((data[17] << 8) | data[16]);
    m_iDig_P7 = (short)((data[19] << 8) | data[18]);
    m_iDig_P8 = (short)((data[21] << 8) | data[20]);
    m_iDig_P9 = (short)((data[23] << 8) | data[22]);

    m_bDig_H1 = (BYTE)(data[24]);                      // 符号なし
    m_iDig_H2 = (short)(data[26] << 8) | data[25];
    m_bDig_H3 = (BYTE)(data[27]);                      // 符号なし
    m_iDig_H4 = (short)((data[28] << 4) | (0x0F & data[29]));
    m_iDig_H5 = (short)((data[30] << 4) | ((data[29] >> 4) & 0x0F));
    m_iDig_H6 = (short)(data[31]);
}